/*
 * This file is part of firk's window manager for x11 (fwm)
 * Copyright (C) 2016-2021 firk <firk@cantconnect.ru>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include "main.h"
#include "config.h"


struct startup_list {
  unsigned int type; /* 0=log, 1=shellcmd */
  char const * cmd;
};


#define MAX_HOTKEYS 200
#define MAX_STARTUP 200
static struct hotkeys_list hotkeys_list[MAX_HOTKEYS];
static size_t num_hotkeys, over_hotkeys;
static struct startup_list startup_list[MAX_STARTUP];
static size_t num_startup, over_startup;
       int startup_errors;

extern struct hotkeys_list * get_hotkeys_list(size_t * count) {
  *count = num_hotkeys;
  return hotkeys_list;
}

extern char const * get_hotkey_by_keycode(unsigned int kc) {
  unsigned int n;
  struct hotkeys_list * hk;
  for(hk=hotkeys_list,n=num_hotkeys; n; hk++,n--) if(hk->keycode==kc) return hk->cmd;
  return NULL;
}

static void add_hotkey(unsigned int keysym, char const * cmd) {
  if(num_hotkeys<MAX_HOTKEYS) {
    hotkeys_list[num_hotkeys].keysym = keysym;
    hotkeys_list[num_hotkeys].cmd = cmd;
    num_hotkeys++;
  } else over_hotkeys++;
}

static void add_startup(unsigned int type, char const * cmd) {
  if(num_startup<MAX_STARTUP) {
    startup_list[num_startup].type = type;
    startup_list[num_startup].cmd = cmd;
    num_startup++;
  } else {
    free((void*)cmd);
    over_startup++;
  }
}

static void set_default_config(void) {
  add_hotkey(XK_Escape, OP_CLOSE_WINDOW);
  add_hotkey(XK_F10,    OP_KILL_WINDOW);
  add_hotkey(XK_Tab,    OP_NEXT_WINDOW);
  add_hotkey(XK_F,      OP_TOGGLE_FULL);
  add_hotkey(XK_grave,  OP_TOGGLE_ONTOP);
  add_hotkey(XK_T,      "gnome-terminal");
  add_hotkey(XK_I,      "iceweasel");
  add_hotkey(XK_P,      "gnome-screenshot -a");
}


static void parse_command(unsigned ln, char * buf) {
  char c;
  char *b0, *b1;
  char const *arg;
  KeySym ks;
  
  while((c=*buf)==' ' || c=='\t') buf++;
  if(c=='#' || !c) return;
  b0 = buf;
  while((c=*buf) && c!=' ' && c!='\t' && c!='#') buf++;
  *buf++ = 0;
  if(c!=' ' && c!='\t') b1 = "";
  else {
    while((c=*buf)==' ' || c=='\t') buf++;
    b1 = buf;
    while((c=*buf) && c!='#') buf++;
    *buf = 0;
  }
  if(!strcmp(b0,"log")) {
    if(!(arg = strdup(b1))) { logprint("out of memory while reading config line %u", ln); startup_errors = 1; return; }
    add_startup(0, arg);
    return;
  }
  if(!strcmp(b0,"run")) {
    if(!(arg = strdup(b1))) { logprint("out of memory while reading config line %u", ln); startup_errors = 1; return; }
    add_startup(1, arg);
    return;
  }
  if(!strncmp(b0,"hotkey:",7)) {
    if((ks = XStringToKeysym(b0+7))==NoSymbol) { logprint("unknown keysym in config line %u", ln); startup_errors = 1; return; }
    if(!strcmp(b1,"OP_CLOSE_WINDOW")) arg = OP_CLOSE_WINDOW;
    else if(!strcmp(b1,"OP_KILL_WINDOW")) arg = OP_KILL_WINDOW;
    else if(!strcmp(b1,"OP_TOGGLE_FULL")) arg = OP_TOGGLE_FULL;
    else if(!strcmp(b1,"OP_NEXT_WINDOW")) arg = OP_NEXT_WINDOW;
    else if(!strcmp(b1,"OP_TOGGLE_ONTOP")) arg = OP_TOGGLE_ONTOP;
    else if(!(arg = strdup(b1))) { logprint("out of memory while reading config line %u", ln); startup_errors = 1; return; }
    add_hotkey(ks, arg);
    return;
  }
  logprint("error in config line %u", ln); startup_errors = 1; return;
}

static int parse_config_fd(int fd) {
  char buf[4000], c;
  unsigned int sz, p, p0, ln;
  int r, skip;
  sz = 0;
  r = 1;
  skip = 0;
  ln = 1;
  while(r) {
    while((r = read(fd, buf+sz, sizeof(buf)-sz))<0) if(errno!=EINTR) return -1;
    for(p=sz,p0=0,sz+=r; p<sz; p++) {
      c = buf[p];
      if(c=='\n') {
        buf[p] = 0;
        if(skip) skip=0; else parse_command(ln, buf+p0);
        p0 = p+1;
        ln++;
      } else if(!skip && (c>=0 && c!=9 && c<32 || c==127)) { logprint("config line %u contains invalid symbols, skipping it", ln); startup_errors = 1; skip = 1; }
    }
    if(p0) if(sz-=p0) bcopy(buf+p0, buf, sz);
  }
  if(sz) { logprint("config line %u (last) non-terminated, skipping it", ln); startup_errors = 1; }
  return 0;
}




static void clear_startup(void) {
  while(num_startup) {
    num_startup--;
    if(startup_list[num_startup].cmd) free((void*)startup_list[num_startup].cmd);
  }
  num_startup = over_startup = 0;
}


static void run_startup_item(unsigned int type, char const * cmd) {
  switch(type) {
  case 0:
    logprint("%s", cmd);
    return;
  case 1:
    system(cmd);
    return;
  }
}


extern void load_config(char const * fn) {
  unsigned int i;
  int fd;
  if(!fn || !*fn) goto defcfg;
  if(fn[0]=='-' && !fn[1]) {
    if(parse_config_fd(0)<0) { logprint("config read(%s) i/o error, falling back to default", fn); goto fallback; }
  } else {
    if((fd = open(fn, O_RDONLY|O_NOCTTY))<0) { logprint("config open(%s) failed, falling back to default", fn); goto fallback; }
    if(parse_config_fd(fd)<0) { logprint("config read(%s) i/o error, falling back to default", fn); close(fd); goto fallback; }
    close(fd);
  }
  if(over_hotkeys) { logprint("too many hotkeys, %u skipped", (unsigned)over_hotkeys); startup_errors = 1; }
  for(i=0; i<num_startup; i++) run_startup_item(startup_list[i].type, startup_list[i].cmd);
  if(over_startup) { logprint("too many startup commands, %u skipped", (unsigned)over_startup); startup_errors = 1; }
  clear_startup();
  return;
fallback:
  num_hotkeys = over_hotkeys = 0; clear_startup(); startup_errors = 1;
defcfg:
  logprint("using default config"); set_default_config(); return;
}
